use snafu::prelude::*;
use qbittorrent_web_api::Api;
use tokio::runtime::{Builder, Runtime};

use crate::Config;
use crate::error::{Error, ApiSnafu as IOError, InternalApiSnafu as ApiError, FailedDeserializeSnafu as DeserializeError};

mod types;
pub use types::QBittorrentPropertiesTorrent;
pub use types::QBittorrentListTorrent;
use crate::api::{Torrent, TorrentList, IntoTorrent};

pub fn blocking_runtime() -> std::io::Result<Runtime> {
    Builder::new_current_thread()
        .enable_all()
        .build()
}

#[derive(Debug)]
pub struct UnauthenticatedRawAsyncApiClient;

impl UnauthenticatedRawAsyncApiClient {
    pub async fn login(host: &str, login: &str, password: &str) -> Result<RawAsyncApiClient, Error> {
        Ok(RawAsyncApiClient {
            api: Api::login(host, login, password).await.context(ApiError)?,
        })
    }
}

#[derive(Debug)]
pub struct RawAsyncApiClient {
    pub api: qbittorrent_web_api::api_impl::Authenticated,
}

impl RawAsyncApiClient {
    #[allow(dead_code)]
    pub async fn add(&self, magnet: &str, paused: bool) -> Result<String, Error> {
        if paused {
            self.api.torrent_management().add(magnet).paused("true").send_raw().await.context(ApiError)
        } else {
            self.api.torrent_management().add(magnet).send_raw().await.context(ApiError)
        }
    }

    #[allow(dead_code)]
    pub async fn get(&self, hash: &str) -> Result<String, Error> {
        self.api.torrent_management().properties_raw(hash).await.context(ApiError)
    }

    #[allow(dead_code)]
    pub async fn list(&self) -> Result<String, Error> {
        self.api.torrent_management().info().send_raw().await.context(ApiError)
    }
}

#[derive(Debug)]
pub struct RawApiClient {
    pub rt: Runtime,
    pub api: RawAsyncApiClient,
}

impl RawApiClient {
    pub fn from_config(config: &Config) -> Result<RawApiClient, Error> {
        Self::new(&config.format_host(), &config.qbittorrent.login, &config.qbittorrent.password)
    }

    /// Login into a qBittorrent backend and return a proper ApiClient instance
    #[allow(dead_code)]
    pub fn new(host: &str, login: &str, password: &str) -> Result<RawApiClient, Error> {
        let rt = blocking_runtime().context(IOError)?;
        let api = rt.block_on(UnauthenticatedRawAsyncApiClient::login(host, login, password))?;
        Ok(RawApiClient {
            rt,
            api,
        })
    }

    #[allow(dead_code)]
    pub fn add(&self, magnet: &str, paused: bool) -> Result<String, Error> {
        self.rt.block_on(self.api.add(magnet, paused))
    }

    #[allow(dead_code)]
    pub fn get(&self, hash: &str) -> Result<String, Error> {
        self.rt.block_on(self.api.get(hash))
    }

    #[allow(dead_code)]
    pub fn list(&self) -> Result<String, Error> {
        self.rt.block_on(self.api.list())
    }
}

#[derive(Debug)]
pub struct AsyncApiClient {
    pub raw_api: RawAsyncApiClient,
}

impl AsyncApiClient {
    #[allow(dead_code)]
    pub async fn from_config(config: &Config) -> Result<AsyncApiClient, Error> {
        Self::new(&config.format_host(), &config.qbittorrent.login, &config.qbittorrent.password).await
    }

    #[allow(dead_code)]
    pub async fn new(host: &str, login: &str, password: &str) -> Result<AsyncApiClient, Error> {
        let api = UnauthenticatedRawAsyncApiClient::login(host, login, password).await?;

        Ok(AsyncApiClient {
            raw_api: api,
        })
    }

    #[allow(dead_code)]
    pub async fn add(&self, magnet: &str, paused: bool) -> Result<(), Error> {
        let res = self.raw_api.add(magnet, paused).await?;
        if res == "Ok." {
            Ok(())
        } else {
            Err(Error::message(res.clone()))
        }
    }

    #[allow(dead_code)]
    pub async fn get(&self, hash: &str) -> Result<Option<Torrent>, Error> {
        let res = self.raw_api.get(hash).await?;
        if res == "" {
            Ok(None)
        } else {
            let concrete: QBittorrentPropertiesTorrent = serde_json::from_str(&res).context(DeserializeError)?;
            Ok(Some(concrete.into_torrent(hash)))
        }
    }

    #[allow(dead_code)]
    pub async fn list(&self) -> Result<TorrentList, Error> {
        let res = self.raw_api.list().await?;
        let concrete: Vec<QBittorrentListTorrent> = serde_json::from_str(&res).context(DeserializeError)?;
        Ok(concrete.iter().map(|t| t.into_torrent(&t.hash)).collect())
    }

}

#[derive(Debug)]
pub struct ApiClient {
    pub rt: Runtime,
    pub api: AsyncApiClient,
}

impl ApiClient {
    #[allow(dead_code)]
    pub fn from_config(config: &Config) -> Result<ApiClient, Error> {
        Self::new(&config.format_host(), &config.qbittorrent.login, &config.qbittorrent.password)
    }

    /// Login into a qBittorrent backend and return a proper ApiClient instance
    #[allow(dead_code)]
    pub fn new(host: &str, login: &str, password: &str) -> Result<ApiClient, Error> {
        let rt = blocking_runtime().context(IOError)?;
        let api = rt.block_on(AsyncApiClient::new(host, login, password))?;
        Ok(ApiClient {
            rt,
            api,
        })
    }

    #[allow(dead_code)]
    pub fn add(&self, magnet: &str, paused: bool) -> Result<(), Error> {
        self.rt.block_on(self.api.add(magnet, paused))
    }

    #[allow(dead_code)]
    pub fn get(&self, hash: &str) -> Result<Option<Torrent>, Error> {
        self.rt.block_on(self.api.get(hash))
    }

    #[allow(dead_code)]
    pub fn list(&self) -> Result<TorrentList, Error> {
        self.rt.block_on(self.api.list())
    }
}
